//
//  ChannelListTableViewController.swift
//  SampleApp
//
//  Created by Federico Zanetello on 10/4/18.
//  Copyright © 2018 David Zhang. All rights reserved.
//

import UIKit
import SwiftUI
import AmitySDK

/*
 * Note:
 *
 * Look into `ChannelListDataSource` class to see the sdk implementation to fetch channel
 * list.
 */
final class ChannelListTableViewController: UIViewController, DataSourceListener, UITableViewDataSource, UITableViewDelegate {
    
    @IBOutlet weak var filterButton: UIButton!
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var updateButton: UIButton!
    @IBOutlet weak var singleChannelButton: UIButton!
    
    // To be injected.
    var repository: AmityChannelRepository! {
        didSet {
            dataSource = ChannelListDataSource(channelRepository: repository)
            dataSource?.dataSourceObserver = self
        }
    }
    
    weak var delegate: ChannelsTableViewControllerDelegate?
    public var channelType: AmityChannelType = .standard
    
    private var dataSource: ChannelListDataSource?
    private lazy var cellConfigurator = ChannelListTableViewCellConfigurator()
    
    let refreshControl = UIRefreshControl()
    
    // MARK: View Lifecycle
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        let nib = UINib(nibName: SampleAppTableViewCell.identifier, bundle: nil)
        tableView.register(nib, forCellReuseIdentifier: SampleAppTableViewCell.identifier)
        
        refreshControl.addTarget(self, action: #selector(fetchChannelList), for: .valueChanged)
        
        tableView.dataSource = self
        tableView.delegate = self
        tableView.refreshControl = refreshControl
        
        filterButton.addTarget(self, action: #selector(onFilterButtonTap), for: .touchUpInside)
        updateButton.addTarget(self, action: #selector(onUpdateButtonTap), for: .touchUpInside)
        singleChannelButton.addTarget(self, action: #selector(onSingleChannelButtonTap), for: .touchUpInside)
    }
    
    @objc func onFilterButtonTap() {
        let filterView = ChannelFilterView {
            self.fetchChannelList()
        }
        let hostingController = UIHostingController(rootView: filterView)
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    @objc func onUpdateButtonTap() {
        let hostingController = UIHostingController(rootView: ChannelUpdateView())
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    @objc func onSingleChannelButtonTap() {
        let hostingController = UIHostingController(rootView: SingleChannelQueryView())
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    /// Fires a new request to fetch all the channels matching the current filters and tags preferences.
    @objc func fetchChannelList() {
        guard let dataSource = dataSource else { return }
        let channelFilter = UserDefaults.standard.filter
        let includingTags = UserDefaults.standard.includingTags
        let excludingTags = UserDefaults.standard.excludingTags
        channelType = UserDefaults.standard.channelTypeFilter

        Log.add(info: "Channel Type: \(UserDefaults.standard.channelTypeFilter.description)")
        Log.add(info: "Filter: \(channelFilter.description)")
        Log.add(info: "Including Tags: \(includingTags), Excluding Tags: \(excludingTags)")
        
        refreshControl.endRefreshing()

        if (channelFilter == .userIsNotMember || channelFilter == .all) && channelType == .private {
            dataSource.fetchChannels(
                channelType: .unknown,
                channelFilter: channelFilter,
                includingTags: includingTags,
                excludingTags: excludingTags,
                channelTypes: Set(arrayLiteral: "private")
            )
        } else {
            dataSource.fetchChannels(
                channelType: channelType,
                channelFilter: channelFilter,
                includingTags: includingTags,
                excludingTags: excludingTags
            )
        }
    }
    
    // MARK: UITableViewDataSource
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return dataSource?.numberOfChannels() ?? 0
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath) else { return UITableViewCell() }
        let cell: UITableViewCell = tableView.dequeueReusableCell(withIdentifier: SampleAppTableViewCell.identifier, for: indexPath)
        guard let sampleAppCell = cell as? SampleAppTableViewCell else { return UITableViewCell() }
        
        cellConfigurator.configure(sampleAppCell, with: channel)
        return sampleAppCell
    }
    
    // MARK: UITableViewDelegate
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath) else { return }
        
        let action1 = UIAlertAction(title: "Join Channel", style: .default) { [weak self] action in
            self?.delegate?.joinChannel(channel: channel)
        }
        
        let action2 = UIAlertAction(title: "View Details", style: .default) { [weak self] action in
            let detailController = UIHostingController(rootView: ChannelDetailsView(detail: ChannelDetail(object: channel)))
            self?.navigationController?.pushViewController(detailController, animated: true)
        }
        
        let action3 = UIAlertAction(title: "Channel Real Time Events", style: .default) { [weak self] action in
            let realTimeController = ChannelRealTimeEventController()
            realTimeController.manager = ChannelRealTimeEventManager(channel: channel)
            
            // start observe only for auto subscription channel
            if channel.channelType == .community || channel.channelType == .conversation {
                realTimeController.shouldStartObserve = true
            }
            
            self?.navigationController?.pushViewController(realTimeController, animated: true)
        }
        
        AppUtility.showActionSheet(in: self, title: "", message: "What would you like to do?", actions: [action1, action2, action3])
    }
    
    func tableView(_ tableView: UITableView, trailingSwipeActionsConfigurationForRowAt indexPath: IndexPath) -> UISwipeActionsConfiguration? {
        let channelId = dataSource?.channel(for: indexPath)?.channelId ?? ""
        let action1 = UIContextualAction(style: .normal, title: "Copy Id") { (action, view, completion) in
            UIPasteboard.general.string = channelId
            completion(true)
        }
        
        let config = UISwipeActionsConfiguration(actions: [action1])
        return config
    }
    
    // MARK: UIScrollViewDelegate
    
    func scrollViewWillEndDragging(_ scrollView: UIScrollView, withVelocity velocity: CGPoint, targetContentOffset: UnsafeMutablePointer<CGPoint>) {
        let height = scrollView.frame.size.height
        let contentYoffset = scrollView.contentOffset.y
        let distanceFromBottom = scrollView.contentSize.height - contentYoffset
        if distanceFromBottom < height {
            dataSource?.fetchMoreChannels()
        }
    }
    
    // MARK: DataSourceListener
    
    func didUpdateDataSource() {
        tableView.reloadData()
    }
}

extension AmityChannelQueryFilter: CustomStringConvertible {
    public var description: String {
        switch self {
        case .all: return "All"
        case .userIsMember: return "User Is Member"
        case .userIsNotMember: return "User Is Not A Member"
        @unknown default:
            assertionFailure("Unknown")
            return "Unknown"
        }
    }
}

extension UIViewController {
    
    func displayActions(actions: [String], title: String, message: String, selectedAction: @escaping (String, Int) -> Void) {
        
        let alertController = UIAlertController(title: title, message: message, preferredStyle: .actionSheet)
        
        for (index, action) in actions.enumerated() {
            let sheetAction = UIAlertAction(title: action, style: .default) { _ in
                selectedAction(action, index)
            }
            alertController.addAction(sheetAction)
        }
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        alertController.addAction(cancelAction)
        
        self.present(alertController, animated: true, completion: nil)
    }
}
